(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import FSets.

(* Based on FSets.FSetFacts from the Coq library.

   We have this signature so that we don't have to keep applying Properties,
   which generates a ton of error messages from Add Morphism. *)

Module Type SetIffType (M:S).

  Import M.

  Module FM.
    
    Section IffSpec.
      Variable s s' s'' : t.
      Variable x y z : elt.
      
      Parameter In_eq_iff : E.eq x y -> (In x s <-> In y s).
      
      Parameter mem_iff : In x s <-> mem x s = true.
      
      Parameter not_mem_iff : ~In x s <-> mem x s = false.
      
      Parameter equal_iff : s[=]s' <-> equal s s' = true.
      
      Parameter subset_iff : s[<=]s' <-> subset s s' = true.
      
      Parameter empty_iff : In x empty <-> False.
      
      Parameter is_empty_iff : Empty s <-> is_empty s = true.
      
      Parameter singleton_iff : In y (singleton x) <-> E.eq x y.
      
      Parameter add_iff : In y (add x s) <-> E.eq x y \/ In y s.
      
      Parameter add_neq_iff : ~ E.eq x y -> (In y (add x s) <-> In y s).
      
      Parameter remove_iff : In y (remove x s) <-> In y s /\ ~E.eq x y.
      
      Parameter remove_neq_iff : ~ E.eq x y -> (In y (remove x s) <-> In y s).
      
      Parameter union_iff : In x (union s s') <-> In x s \/ In x s'.
      
      Parameter inter_iff : In x (inter s s') <-> In x s /\ In x s'.
      
      Parameter diff_iff : In x (diff s s') <-> In x s /\ ~ In x s'.
      
      Variable f : elt->bool.
      
      Parameter filter_iff :
        compat_bool E.eq f -> (In x (filter f s) <-> In x s /\ f x = true).
      
      Parameter for_all_iff : compat_bool E.eq f ->
        (For_all (fun x => f x = true) s <-> for_all f s = true).
      
      Parameter exists_iff : compat_bool E.eq f ->
        (Exists (fun x => f x = true) s <-> exists_ f s = true).

      Parameter elements_iff : In x s <-> InA E.eq x (elements s).

    End IffSpec.

  End FM.

  (* adding in other things from FSetProperties, ad hoc.  *)
  Parameter add_cardinal_2 :
   forall s x, ~In x s -> cardinal (add x s) = S (cardinal s).

  Parameter cardinal_1 : forall s, Empty s -> cardinal s = 0.

  Parameter add_remove : forall s x, In x s -> add x (remove x s) [=] s.

  Parameter remove_add : forall s x, ~In x s -> remove x (add x s) [=] s.

  Parameter empty_is_empty_1 : forall s, Empty s -> s[=]empty.

  Parameter empty_is_empty_2 : forall s, s[=]empty -> Empty s.

  Definition Add (x : elt) (s s' : t) :=
    forall y : elt, In y s' <-> E.eq x y \/ In y s.

  Parameter set_induction :
    forall P : t -> Type,
      (forall s : t, Empty s -> P s) ->
      (forall s s' : t, P s -> forall x : elt, ~In x s -> Add x s s' -> P s') ->
      forall s : t, P s.

End SetIffType.