(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import List.


Ltac elf := elimtype False.

(* somewhat obsoleted by apply in? *)
Ltac hyr H H' :=
  let H'' := fresh in
    (assert (H'' := H'); clear H; rename H'' into H).

Ltac injector E :=
  injection E; clear E; intros; subst.

Ltac intuit := intuition.



Tactic Notation "computet" constr(X) "in" hyp(H) :=
  let Z := fresh in
    set (Z := X) in H;
    compute in Z;
    subst Z.

Tactic Notation "computet" constr(X) :=
  let Z := fresh in
    set (Z := X);
    compute in Z;
    subst Z.


Tactic Notation "vm_computet" constr(X) "in" hyp(H) :=
  let Z := fresh in
    set (Z := X) in H;
    vm_compute in Z;
    subst Z.

Tactic Notation "vm_computet" constr(X) :=
  let Z := fresh in
    set (Z := X);
    vm_compute in Z;
    subst Z.


Ltac autof := auto; fail.

Ltac nameSwap x y :=
  let temp := fresh in
    rename x into temp;
      rename y into x;
        rename temp into y.



(* adding some additional rewriting rules for lists *)

Lemma lengthNilEq : forall A,
  length (@nil A) = 0%nat.
Proof.
  reflexivity.
Qed.

Lemma lengthConsEq : forall A (x:A) l,
  length (cons x l) = S (length l).
Proof.
  reflexivity.
Qed.

Hint Rewrite lengthNilEq lengthConsEq app_ass : list.
