(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import ProgLog.
Require Import NatUtil.


Set Implicit Arguments.


Definition xv := Var 0.
Definition yv := Var 1.
Definition tempv := Var 2.

(* Swap
   Arguments: xv and yv, both pointers
   The function swaps the vals contained in these pointers.
*)
Notation swapBody :=
  (tempv <- [xv];;
   [xv]  <- [yv];;
   [yv]  <- tempv;;
   ret 0)%CM.

                                (* stack space*)
Notation swapDef := 
    (fdefi (xv::yv::nil) (tempv::nil) 0 swapBody).
          (* args*)     (* locals *)   (* body *)

Notation swapTy := (val::val::nil : tslist).
  (* types of the auxiliary variables *)

Definition swapPre a b i j :=
  i |-> a ** j |-> b.

Definition swapPost a b i j (v:val) :=
  i |-> b ** j |-> a.

Definition P0 : cdhpty := fun _ => None.



Lemma swapOk : fdefOk P0 swapTy 2 swapPre swapDef swapPost.
Proof.
  fdefBegin.
  unfold swapPre, swapPost. intros a b i j m sp Hp vf VFE.
  ssimpl in Hp. simpl in VFE.
  vcSteps.
  ssimpl.
Qed.


(* An alternate proof script that shows what happens more step by
step. *)
Lemma swapOk2 : fdefOk P0 swapTy 2 swapPre swapDef swapPost.
Proof.
  fdefBegin.
  unfold swapPre, swapPost. intros a b i j m sp Hp vf VFE.
  ssimpl in Hp. simpl in VFE.

  (* First statement   temp <- [xv]  *)
  basicStmVCStep. (* extract the first statement *)
     (* first eval the load, then the rest *)
  basicStmVCStep. (* unpack eval of the first statement *)
     (* eval the RHS, update the value *)

  autorewrite with ProgLog.     (* minor simplification *)
  autorewrite with ProgLogExp.  (* unpack eval of the RHS *)
    (* eval xv, load the value *)

  fancyExpStep.  (* xv has value i, from VFE *)
  fancyExpStep.  (* address i has value a, from Hp *)
  autorewrite with ProgLog. (* minor simplification *)
  fancyExpStep.  (* update tempv to have value a, new VFE *)


  (* Second statement   [xv] <- [yv] *)
  basicStmVCStep. (* extract the second statement *)  
  basicStmVCStep. (* unpack eval of the second statement *)
     (* eval the LHS, eval the RHS, store new value *)

  autorewrite with ProgLogExp.  (* unpack eval of the RHS *)
    (* eval xv, load its value *)

  fancyExpStep.  (* xv has value i, from VFE *)
  fancyExpStep.  (* yv has value j, from VFE *)
  autorewrite with ProgLog. (* minor simplification *)

  fancyExpStep.  (* address j has value a, from Hp *)
  autorewrite with ProgLog. (* minor simplification *)
  fancyExpStep.  (* update addr i to value b, new Hp *)


  (* Third statement   [yv] <- tempv *)
  basicStmVCStep. (* extract the third statement *)  
  basicStmVCStep. (* unpack eval of the third statement *)
     (* eval the LHS, eval the RHS, store new value *)

  fancyExpStep.  (* yv has value j, from VFE *)
  fancyExpStep.  (* xv has value i, from VFE *)
  autorewrite with ProgLog. (* minor simplification *)
  fancyExpStep.  (* update addr j to value a, new Hp *)

  
  (* Final statement    ret 0 *)
  basicStmVCStep.   (* unpack eval of the return *)
    (* also select the final thing we need to prove out of the VC args *)
  autorewrite with ProgLogExp ProgLog. (* eval is trivial *)

  (* We have now shown that the statement body can be executed safely.
     All that remains is to show that the post condition holds. *)
  unfold swapPost.  (* unfold the post-condition *)

  (* commutativity, get rid of empty stack frame, then it matches Hp *)
  ssimpl.

Qed.






